/*
 *  (c) 2005,2006 Frank A. Zdarsky, disco | Distributed Computer Systems Lab
 *                                  University of Kaiserslautern, Germany
 *         All Rights Reserved.
 *
 *  This software is work in progress and is released in the hope that it will
 *  be useful to the scientific community. It is provided "as is" without
 *  express or implied warranty, including but not limited to the correctness
 *  of the code or its suitability for any particular purpose.
 *
 *  You are free to use this software for any non-commercial educational or
 *  research purpose, provided that this copyright notice is not removed or
 *  modified. For commercial uses please contact the respective author(s).
 *
 *  If you find our software useful, we would appreciate if you mentioned it
 *  in any publication arising from the use of this software or acknowledge
 *  our work otherwise. We would also like to hear of any fixes or useful
 *  extensions to this software.
 *
 */

package unikl.disco.misc;

import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import edu.uci.ics.jung.graph.DirectedEdge;
import edu.uci.ics.jung.graph.Vertex;
import edu.uci.ics.jung.graph.impl.DirectedSparseEdge;
import edu.uci.ics.jung.graph.impl.DirectedSparseGraph;
import edu.uci.ics.jung.graph.impl.SimpleSparseVertex;

/**
 * A tiny collection of methods useful in dealing with graphs but not provided by JUNG.
 * 
 * @author Frank A. Zdarsky
 */
public class GraphUtils {
	/**
	 * Returns the subsets of all nodes of the directed graph <code>g</code> that
	 * are sources, i.e. whose predecessor set is empty (in-degree == 0).
	 * @param g the graph
	 * @return the set of source nodes
	 */
	public static Set getSourceSet(DirectedSparseGraph g) {
		Set sources = new HashSet();
		for (Iterator iter = g.getVertices().iterator(); iter.hasNext(); ) {
			SimpleSparseVertex v = (SimpleSparseVertex) iter.next();
			if (v.inDegree() == 0) {
				sources.add(v);
			}
		}
		return sources;
	}
	
	/**
	 * Returns the subsets of all nodes of the directed graph <code>g</code> that
	 * are sinks, i.e. whose successor set is empty (out-degree == 0).
	 * @param g the graph
	 * @return the set of sink nodes
	 */
	public static Set getSinkSet(DirectedSparseGraph g) {
		Set sinks = new HashSet();
		for (Iterator iter = g.getVertices().iterator(); iter.hasNext(); ) {
			SimpleSparseVertex v = (SimpleSparseVertex) iter.next();
			if (v.outDegree() == 0) {
				sinks.add(v);
			}
		}
		return sinks;
	}
	
	/**
	 * Returns the subsets of all nodes of the directed graph <code>g</code> that
	 * are neither sources nor sinks, i.e. who have both (a) predecessor(s) and
	 * (a) successor(s) (in-degree > 0 and out-degree > 0).
	 * @param g the graph
	 * @return the set of server nodes
	 */
	public static Set getServerSet(DirectedSparseGraph g) {
		Set servers = new HashSet();
		for (Iterator iter = g.getVertices().iterator(); iter.hasNext(); ) {
			SimpleSparseVertex v = (SimpleSparseVertex) iter.next();
			if (v.inDegree() > 0 && v.outDegree() > 0) {
				servers.add(v);
			}
		}
		return servers;
	}
	
    /**
     * Returns a sublist view of the given edge path <code>path</code> from the vertex
     * <code>from</code> to vertex <code>to</code> or an empty list if the vertices are not
     * both on the path and traversed in order.
     * @param path
     * @param from
     * @param to
     * @return the sub path from <code>from</code> to <code>to</code>
     */
    public static List getSubPath(List path, Vertex from, Vertex to) {
		int iFrom = -1;		
		int i     = 0;
		for (Iterator iter = path.iterator(); iter.hasNext(); i++ ) {
			DirectedEdge e = (DirectedEdge) iter.next();
			if (e.getSource().equals(from)) {
				iFrom = i;
			}
			if (e.getDest().equals(to)) {
				return (iFrom >= 0) ? path.subList(iFrom, i+1) : Collections.EMPTY_LIST;
			}
		}
		return Collections.EMPTY_LIST;
	}
    
    /**
     * Creates and returns server graph from a given <code>network_graph</code>
     * by creating a server for each link and an edge between servers for each
     * turn around a router.<br/>
     * If empty map instances are passed for
     * <code>link_server_map</code> and/or <code>router_turns_map</code>, the maps
     * will be filled with mappings of links (instance of <code>DirectedEdge</code>)
     * to servers (instance of <code>SimpleSparseVertex</code> and routers
     * (instance of <code>SimpleSparseVertex</code>) to a set of turns
     * (instance of <code>DirectedEdge</code>). If you do not need these mappings,
     * simply pass <code>null</code> instead.<br/>
     * @param network_graph the network graph
     * @param link_server_map a mapping of links to servers
     * @param router_turns_map a mapping of routers to sets of turns (= edge between two servers)
     * @return a server graph
     */
    public static DirectedSparseGraph createServerGraph(DirectedSparseGraph network_graph,
    		Map link_server_map, Map router_turns_map) {
		DirectedSparseGraph server_graph = new DirectedSparseGraph();
		
		Map internal_link_server_map  = (link_server_map != null) ?
				link_server_map : new HashMap();
		Map internal_router_turns_map = (router_turns_map != null) ?
				router_turns_map : new HashMap();
		
		// create a server for each link
		for (Iterator iter = network_graph.getEdges().iterator(); iter.hasNext(); ) {
			DirectedEdge link = (DirectedEdge) iter.next();
			
			Vertex server = new SimpleSparseVertex();
			server_graph.addVertex(server);
			internal_link_server_map.put(link, server);
		}
		
		// create a turn between two servers for each incoming-outgoing link pair of routers
		for (Iterator inIter = network_graph.getEdges().iterator(); inIter.hasNext(); ) {
			DirectedEdge inlink = (DirectedEdge) inIter.next();
			
			Vertex router = (Vertex) inlink.getEndpoints().getSecond();
			if (!internal_router_turns_map.containsKey(router)) {
				internal_router_turns_map.put(router, new HashSet());
			}

			for (Iterator outIter = router.getOutEdges().iterator(); outIter.hasNext(); ) {
				DirectedEdge outlink = (DirectedEdge) outIter.next();
				
				if (inlink.getSource() == outlink.getDest()) {
					continue;
				}
				
				DirectedEdge turn = new DirectedSparseEdge(
						(Vertex) internal_link_server_map.get(inlink),
						(Vertex) internal_link_server_map.get(outlink));
				server_graph.addEdge(turn);
				((Set) internal_router_turns_map.get(router)).add(turn);
			}
		}
		
		return server_graph;
    }
    
    /**
     * Creates and returns server graph by traversing all elements in
     * the list of paths <code>pahts</code> and creating a server for each
     * each traversed link (if none exists yet) and an edge between servers
     * for each pair of traversed links (if none exists yet.<br/>
     * If empty map instances are passed for
     * <code>link_server_map</code> and/or <code>router_turns_map</code>, the maps
     * will be filled with mappings of links (instance of <code>DirectedEdge</code>)
     * to servers (instance of <code>SimpleSparseVertex</code> and routers
     * (instance of <code>SimpleSparseVertex</code>) to a set of turns
     * (instance of <code>DirectedEdge</code>). If you do not need these mappings,
     * simply pass <code>null</code> instead.<br/>
     * @param paths a list of edge paths
     * @param link_server_map a mapping of links to servers
     * @param router_turns_map a mapping of routers to sets of turns (= edge between two servers)
     * @return a server graph
     */
    public static DirectedSparseGraph createServerGraphFromPaths(List paths,
    		Map link_server_map, Map router_turns_map) {
		DirectedSparseGraph server_graph = new DirectedSparseGraph();
		
		Map internal_link_server_map  = (link_server_map != null) ?
				link_server_map : new HashMap();
		Map internal_router_turns_map = (router_turns_map != null) ?
				router_turns_map : new HashMap();
		
		for (Iterator pIter = paths.iterator(); pIter.hasNext(); ) {
			List path = (List) pIter.next();
			
			Vertex previous_server = null;
			for (Iterator iter = path.iterator(); iter.hasNext(); ) {
				DirectedEdge link = (DirectedEdge) iter.next();
			
				Vertex server;
				if (!internal_link_server_map.containsKey(link)) {
					server = new SimpleSparseVertex();
					server_graph.addVertex(server);
					internal_link_server_map.put(link, server);				
				} else {
					server = (Vertex) internal_link_server_map.get(link);
				}
				if (previous_server != null && previous_server.findEdge(server) == null) {
					DirectedEdge turn = new DirectedSparseEdge(previous_server, server);
					server_graph.addEdge(turn);
					
					Vertex router = (Vertex) link.getEndpoints().getFirst();
					if (!internal_router_turns_map.containsKey(router)) {
						internal_router_turns_map.put(router, new HashSet());
					}
					((Set) internal_router_turns_map.get(router)).add(turn);				
				}
				previous_server = server;
			}
		}		
		return server_graph;
    }
}
