package unikl.disco.nc;

import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.ArrayList;
import java.util.Map;
import java.util.HashMap;
import java.util.Set;
import java.util.HashSet;

import edu.uci.ics.jung.graph.DirectedGraph;
import edu.uci.ics.jung.graph.Vertex;
import edu.uci.ics.jung.graph.DirectedEdge;
import edu.uci.ics.jung.utils.UserData;

import unikl.disco.misc.GraphUtils;
import unikl.disco.misc.SetUtils;
import unikl.disco.misc.MapValueComparator;

/**
 * Class containing various methods to analyse a network.
 * It operates on a server graph (i.e. vertices representing servers, edges representing
 * dependencies between servers) which must not contain cyclic dependencies.
 * 
 * @author Frank A. Zdarsky
 */
public class NetworkAnalyser {
	/** Whether to use FIFO service instead of blind multiplexing */
	protected boolean use_fifo_service;
	/** Whether to use maximum service curves in output bound computation */
	protected boolean use_gamma;
	/** Whether to constrain the output bound further through convolution with the maximum service curve */
	protected boolean use_extra_gamma;
	/**
	 * How many levels of recursion the tighter <code>computePMOOOutputBound()</code> is used
	 * rather than standard <code>computeOutputBound()</code>.
	 */
	protected int max_rec_depth_for_pmoo;
	/** Whether to use arrival curve approximation */
	protected boolean use_arrival_curve_approximation;
	/** The targeted number of TBs per flow when reducing complexity */
	protected double  token_bucket_count_target;
	/** Whether to prolong flows to reduce complexity */
	protected boolean use_flow_prolongation;

	/** The server model graph to analyse */
	protected DirectedGraph server_graph;
	/** The current recursion level */
	protected int rec_depth;
	/** A list of all flows in the network */
	protected List flows;
	
	/**
	 * @param server_graph the (server) graph to analyse
	 */
	public NetworkAnalyser(DirectedGraph server_graph) {
		this.server_graph = server_graph;
		this.flows = new ArrayList();
		
		this.use_fifo_service = false;
		this.use_gamma = false;
		this.use_extra_gamma = false;
		this.max_rec_depth_for_pmoo = Integer.MAX_VALUE;
		this.use_arrival_curve_approximation = false;
		this.token_bucket_count_target = 2.0;
		this.use_flow_prolongation = false;
	}
	
	public boolean useFifoService() {
		return use_fifo_service;
	}
	public void setUseFifoService(boolean use_fifo_service) {
		this.use_fifo_service = use_fifo_service;
	}
	public boolean useGamma() {
		return use_gamma;
	}
	public void setUseGamma(boolean use_gamma) {
		this.use_gamma = use_gamma;
	}
	public boolean useExtraGamma() {
		return use_extra_gamma;
	}
	public void setUseExtraGamma(boolean use_extra_gamma) {
		this.use_extra_gamma = use_extra_gamma;
	}
    public int getMaxRecursionDepthForPMOO() {
		return max_rec_depth_for_pmoo;
	}
	public void setMaxRecursionDepthForPMOO(int max_rec_depth_for_pmoo) {
		this.max_rec_depth_for_pmoo = max_rec_depth_for_pmoo;
	}
	public boolean useFlowProlongation() {
		return use_flow_prolongation;
	}
	public void setUseFlowProlongation(boolean use_flow_prolongation) {
		this.use_flow_prolongation = use_flow_prolongation;
	}
	public boolean useArrivalCurveApproximation() {
		return use_arrival_curve_approximation;
	}
	public void setUseArrivalCurveApproximation(boolean use_arrival_curve_approximation) {
		this.use_arrival_curve_approximation = use_arrival_curve_approximation;
	}
	public double getTokenBucketCountTarget() {
		return token_bucket_count_target;
	}
	public void setTokenBucketCountTarget(double token_bucket_count_target) {
		this.token_bucket_count_target = token_bucket_count_target;
	}

	//---------------------------------------------------------------------
    // service curve operations
    //---------------------------------------------------------------------
	/**
	 * Returns the (minimum) service curve of a given <code>server<code> or
	 * a null-curve if no service curve is defined for the server.
	 * @param server
	 * @return the service curve
	 */
	public Curve getServiceCurve(Vertex server) {
		Curve beta = (Curve) server.getUserDatum("service_curve");
		return (beta != null) ? beta : Curve.NULL_CURVE;
	}

	/**
	 * Sets the service curve for the specified server.
	 * @param server the server
	 * @param service_curve the service curve
	 */
	public void setServiceCurve(Vertex server, Curve service_curve) {
		server.setUserDatum("service_curve", service_curve, UserData.REMOVE);
	}
	
	/**
	 * Sets the service curve for the specified set of servers to a copy of
	 * the specified <code>prototype_service_curve</code>
	 * @param servers the set of servers whose service curve shall be set
	 * @param prototype_service_curve the prototype service curve
	 */
	public void setServiceCurve(Set servers, Curve prototype_service_curve) {
		for (Iterator iter = servers.iterator(); iter.hasNext(); ) {
			Vertex server = (Vertex) iter.next();
			server.setUserDatum("service_curve", prototype_service_curve.copy(), UserData.REMOVE);
		}
	}
	
	/**
	 * Returns the maximum service curve of a given <code>server</code> or a
	 * burst-delay-curve with delay 0.0 if no maximum service curve is defined
	 * for the server.
	 * @param server
	 * @return the maximum service curve
	 */
	public Curve getMaxServiceCurve(Vertex server) {
		Curve gamma = (Curve) server.getUserDatum("max_service_curve");
		return (gamma != null) ? gamma : Curve.ZERO_DELAY_BURST_CURVE;
	}

	/**
	 * Sets the maximum service curve for the specified server.
	 * @param server the server
	 * @param max_service_curve the service curve
	 */
	public void setMaxServiceCurve(Vertex server, Curve max_service_curve) {
		server.setUserDatum("max_service_curve", max_service_curve, UserData.REMOVE);		
	}
	
	/**
	 * Sets the maximum service curve for the specified set of servers to a copy of
	 * the specified <code>prototype_max_service_curve</code>
	 * @param servers the set of servers whose maximum service curve shall be set
	 * @param prototype_max_service_curve the prototype maximum service curve
	 */
	public void setMaxServiceCurve(Set servers, Curve prototype_max_service_curve) {
		for (Iterator iter = servers.iterator(); iter.hasNext(); ) {
			Vertex server = (Vertex) iter.next();
			server.setUserDatum("max_service_curve", prototype_max_service_curve.copy(), UserData.REMOVE);
		}
	}
	
	/**
	 * Adjusts the rates of a given server's service curves so that the utilization
	 * of the server is below or equal to <code>target_link_utilization</code>. <br />
	 * More precisely, if the ratio between the sum of the sustained rates of all
	 * flows crossing the server and the server's minimum service rate is higher than
	 * <code>target_link_utilization</code>, the mimimum service rate is increased
	 * until the target utilization is reached. Furthermore, the rate of the maximum
	 * service curve is increased as well, so that the ratio between maximum rate
	 * and minimum rate is preserved.
	 * 
	 * @param server
	 * @param target_link_utilization the maximum link utilization for the server
	 */
    public void autoAdjustServerRate(Vertex server, double target_link_utilization) {
   		double sum_R = 0;
   		for (Iterator flow_iter = getAllIncomingFlows(server).iterator(); flow_iter.hasNext(); ) {
   			sum_R += ((Flow) flow_iter.next()).ac.getSustainedRate();
   		}
    		
   		Curve beta  = (Curve) server.getUserDatum("service_curve");
   		Curve gamma = (Curve) server.getUserDatum("max_service_curve");
    		
   		double C     = beta.getSustainedRate();
   		double C_max = gamma.getSustainedRate();
   		double ratio = C_max/C;
   		C     = Math.max(C, sum_R/target_link_utilization);
   		C_max = Math.max(C, sum_R/target_link_utilization) * ratio;
   		beta.setSustainedRate(C);
   		gamma.setSustainedRate(C_max);

   		server.setUserDatum("service_curve", beta, UserData.REMOVE);
   		server.setUserDatum("max_service_curve", gamma, UserData.REMOVE);
    }
    
    /**
     * Returns the convolution of the service curves of all servers on the
     * given edge path <code>path</code>.<br />
     * All servers having outgoing edges contained in the given edge path
     * (i.e. all but the server at which the path ends) are included.
     * @param path an edge path
     * @return the convolved curve
     */
    public Curve convolveServiceCurves(List path) {
		Curve beta_total = Curve.ZERO_DELAY_BURST_CURVE;
		for (Iterator path_iter = path.iterator(); path_iter.hasNext(); ) {
			DirectedEdge e = (DirectedEdge) path_iter.next();
			Vertex server = e.getSource();
			beta_total = beta_total.convolveConvex(getServiceCurve(server));
		}
		return beta_total;
    }
       
    /**
     * Returns the convolution of the maximum service curves of all servers on the 
     * given edge path <code>path</code>.<br />
     * All servers having outgoing edges contained in the given edge path
     * (i.e. all but the server at which the path ends) are included.
     * @param path an edge path
     * @return the convolved curve
     */
	public Curve convolveMaxServiceCurves(List path) {
		Curve gamma_total = Curve.ZERO_DELAY_BURST_CURVE;
		for (Iterator path_iter = path.iterator(); path_iter.hasNext(); ) {
			DirectedEdge e = (DirectedEdge) path_iter.next();
			Vertex server = e.getSource();
   			gamma_total = gamma_total.convolveAlmostConcave(getMaxServiceCurve(server));
    	}
		return gamma_total;
	}

	//---------------------------------------------------------------------
    // flow operations
    //---------------------------------------------------------------------
	/**
	 * Adds a flow to the network analyser. The flow instance has to be
	 * fully initialized, including the flow's path.
	 * @param flow the flow to be added
	 */
    public void addFlow(Flow flow) {
    	if (!flow.source.containsUserDatumKey("source_flow_set")) {
    		flow.source.addUserDatum("source_flow_set", new HashSet(), UserData.REMOVE);
    	}
		((Set) flow.source.getUserDatum("source_flow_set")).add(flow);
		flows.add(flow);

		for (Iterator edge_iter = flow.path.iterator(); edge_iter.hasNext(); ) {
			DirectedEdge e = (DirectedEdge) edge_iter.next();
			Set flow_set = (Set) e.getUserDatum("flow_set");
			if (flow_set == null) {
				flow_set = new HashSet();
				e.setUserDatum("flow_set", flow_set, UserData.REMOVE);
			}
			flow_set.add(flow);
		}
	}
    
    /**
     * Removes a flow from the network analyser. 
     * @param flow the flow to be removed
     */
	public void removeFlow(Flow flow) {
		for (Iterator edge_iter = flow.path.iterator(); edge_iter.hasNext(); ) {
			DirectedEdge e = (DirectedEdge) edge_iter.next();
			getFlowSet(e).remove(flow);
		}
		
		flows.remove(flow);
    	if (flow.source.containsUserDatumKey("source_flow_set")) {
    		Set flows = (Set) flow.source.getUserDatum("source_flow_set");
    		flows.remove(flow);
    		if (flows.size() == 0) {
    			flow.source.removeUserDatum("source_flow_set");
    		}
    	}
	}
	
	/**
	 * Returns all flows registered with the network analyser.
	 * @return a list of flows
	 */
	public List getFlows() {
		return flows;
	}
   
	/**
	 * Returns a set containing all flows traversing the given edge <code>e</code>.
	 * @param e the traversed edge
	 * @return a set of traversing flows
	 */
    public Set getFlowSet(DirectedEdge e) {
    	Set flow_set = (Set) e.getUserDatum("flow_set");
    	return (flow_set != null) ? flow_set : Collections.EMPTY_SET;
    }
    
    /**
     * Returns a set containing all flows traversing the edge from
     * <code>predecessor</code> to <code>server</code>.
     * @param predecessor
     * @param server
     * @return a set of incoming flows
     */
    public Set getIncomingFlows(Vertex predecessor, Vertex server) {
    	return getFlowSet((DirectedEdge) predecessor.findEdge(server));
    }
    
    /**
     * Returns the set of all flows that enter <code>server</code> from one
     * of its predecessors.
     * @param server
     * @return a set of all incoming flows
     */
	public Set getAllIncomingFlows(Vertex server) {
		Set all_flows = new HashSet();
		for (Iterator inedge_iter = server.getInEdges().iterator(); inedge_iter.hasNext(); ) {
			DirectedEdge e = (DirectedEdge) inedge_iter.next();
			all_flows.addAll(getFlowSet(e));
		}
		return all_flows;
	}

	/**
	 * Returns an aggregate arrival curve for all flows originating in
	 * <code>source</code> and contained in the set <code>outgoing_flows</code>.
	 * @param source the source of all flows to be aggregated
	 * @param outgoing_flows the set of all flows to be aggregated
	 * @return an aggregate arrival curve
	 */
	public Curve getSourceFlow(Vertex source, Set outgoing_flows) {
		Curve a_out = Curve.NULL_CURVE;
		Set flows = (Set) source.getUserDatum("source_flow_set");
		if (flows != null) {
			for (Iterator iter = SetUtils.getIntersection(flows, outgoing_flows).iterator(); iter.hasNext(); ) {
				Flow f = (Flow) iter.next();
				a_out = a_out.add(f.ac);
			}
		}
		return a_out;
	}
	    
	/**
	 * Returns the predecessor server from which a given <code>server</code> receives
	 * the flow <code>f</code>.
	 * @param server the receiving server
	 * @param f the incoming flow
	 * @return the predecessor
	 */
    public Vertex getPredecessor(Vertex server, Flow f) {
    	for (Iterator edge_iter = server.getInEdges().iterator(); edge_iter.hasNext(); ) {
    		DirectedEdge e = (DirectedEdge) edge_iter.next();
    		if (getFlowSet(e).contains(f)) {
    			return e.getSource();
    		}
    	}
    	return null;
    }
    
    /**
     * Returns the length of the longest flow in the given list of flows.
     * @param flows a list of flows
     * @return the length of the longest flow
     */
    public int getMaxFlowLength(List flows) {
    	int length = 0;
    	for (Iterator iter = flows.iterator(); iter.hasNext(); ) {
    		Flow f = (Flow) iter.next();
    		length = (f.path.size() > length) ? f.path.size() : length;
    	}
    	return length;
    }

    //-------------------------------------------------------------------------
    // Charny bound
    //-------------------------------------------------------------------------
    /**
     * Returns the maximum server utilization for which the Charny bound exist.<br />
     * The Charny bound makes the following assumptions:
     * <ul>
     * <li>All fresh arrival curves have token bucket form.</li>
     * <li>All (minimum) service curves have rate latency form.</li>
     * <li>Within an aggregate, packets are served in a FIFO manner.</li>
     * <li>If maximum service curves exist they are of peak rate form.</li>
     * <li>Sources do not send with a higher rate than the maximum rate of any
     * server in the network.</li>
     * </ul>
     */
    public double getMaxUtilizationForCharnyBound() {
    	int h = getMaxFlowLength(flows) - 1;
    	double C_max = getMaxServerRate();
    	double utilization = 1.0;
    	for (Iterator iter = server_graph.getVertices().iterator(); iter.hasNext(); ) {
    		Vertex server = (Vertex) iter.next();
    		if (server.inDegree() == 0 || server.outDegree() == 0) {
    			continue;
    		}
    		
    		double C_m = getIncomingFlowPeakRateBound(server, C_max);
    		double r_m = getServiceCurve(server).getSustainedRate();	
    		if (C_m >= r_m) {
    			utilization = Math.min(utilization, C_m / ((C_m-r_m)*(h-1) + r_m));
    		}
    	}
    	return utilization;
    }

    /**
     * Returns the Charny delay bound if it exists, otherwise returns
     * <code>Double.POSITIVE_INFINITY</code>.
     * @return the delay bound
     * @see NetworkAnalyser#getMaxUtilizationForCharnyBound()
     */
    public double getCharnyDelayBound() {
    	double ny  = getHighestServerUtilization();
    	if (ny >= getMaxUtilizationForCharnyBound()) {
    		return Double.POSITIVE_INFINITY;
    	}
    	int    h   = getMaxFlowLength(flows) - 1;
    	double E   = getHighestServerLatency();
    	double tau = getMaximumPacketDelayVariation();
    	double u   = computeHighestU(ny);
    	return h * (E + u*tau*ny) / (1 - (h-1)*u*ny);
    }
    
    /**
     * Computes the Charny backlog bound for each server. If the bound
     * exists, returns a map containing the servers (<code>Vertex</code> instances)
     * as keys and their corresponding backlog bound (<code>Double</code> instances)
     * as values. If the bound does not exist, returns an empty map.
     * @return a map from servers to backlog bounds
     * @see NetworkAnalyser#getMaxUtilizationForCharnyBound()
     */
    public Map getCharnyBacklogBounds() {
    	Map server_bound_map = new HashMap();
    	
    	double ny  = getHighestServerUtilization();
    	if (ny >= getMaxUtilizationForCharnyBound()) {
    		return server_bound_map; 
    	}

    	int    h   = getMaxFlowLength(flows) - 1;
   		double E   = getHighestServerLatency();
   		double tau = getMaximumPacketDelayVariation();
   		double u   = computeHighestU(ny);
   		double D1  = (E + u*tau*ny) / (1 - (h-1)*u*ny);

    	for (Iterator iter = server_graph.getVertices().iterator(); iter.hasNext(); ) {
    		Vertex server = (Vertex) iter.next();
    		if (server.inDegree() == 0 || server.outDegree() == 0) {
    			continue;
    		}
    		
    		double r_m   = getServiceCurve(server).getSustainedRate();
    		double bound = r_m * ny * (E + tau + D1 * (h-1))
    		        - r_m * Math.max(0.0, ny * (E + tau) - D1*(1 - (h-1)*tau));
    		server_bound_map.put(server, new Double(bound));
    	}
    	return server_bound_map;
    }

    /**
     * Returns the highest server utilization in the network. For the existence of
     * the Charny bounds this utilization must be smaller than
     * <code>getMaxUtilizationForCharnyBound()</code>.
     * @return the highest server utilization
     * @see NetworkAnalyser#getMaxUtilizationForCharnyBound()
     */
    public double getHighestServerUtilization() {
    	double ny = 0.0;
    	for (Iterator node_iter = server_graph.getVertices().iterator(); node_iter.hasNext(); ) {
    		Vertex server = (Vertex) node_iter.next();
    		if (server.inDegree() == 0 || server.outDegree() == 0) {
    			continue;
    		}

    		double r = 0.0;
        	for (Iterator iter = getAllIncomingFlows(server).iterator(); iter.hasNext(); ) {
        		r += ((Flow) iter.next()).ac.getSustainedRate();
        	}
    		ny = Math.max(ny, r / getServiceCurve(server).getSustainedRate());
    	}
    	return ny;    	
    }
    
    /** returns the maximum server rate of any server in the network */
    private double getMaxServerRate() {
    	double C_max = 0.0;
		for (Iterator iter = server_graph.getVertices().iterator(); iter.hasNext(); ) {
			Vertex node = (Vertex) iter.next();
			if (node.inDegree() > 0 && node.outDegree() > 0) {
				Curve gamma = getMaxServiceCurve(node);
    			if (gamma.isBurstDelay() && gamma.getLatency() == 0.0) {
    				return Double.POSITIVE_INFINITY;
    			}
    			C_max = Math.max(C_max, gamma.getHighestRate());
			}
		}
    	return C_max;
    }
    
    /** computes the value C_m in [LeBoudec91] p. 111 */
    private double getIncomingFlowPeakRateBound(Vertex server, double C_max) {
		double C_m   = 0.0;
		int source_count = 0;
		for (Iterator pred_iter = server.getPredecessors().iterator(); pred_iter.hasNext(); ) {
			Vertex pred = (Vertex) pred_iter.next();
			if (pred.inDegree() > 0) {
				Curve gamma = getMaxServiceCurve(pred);
    			if (gamma.isBurstDelay() && gamma.getLatency() == 0.0) {
    				return Double.POSITIVE_INFINITY;
    			}
    			C_m  += gamma.getHighestRate();
			}
			else {
				source_count++;
			}
		}
		C_m += source_count * C_max;
		return C_m;
    }
    
    /** computes the value E in [LeBoudec91] p. 111 */
    private double getHighestServerLatency() {
    	double E = 0.0;
    	for (Iterator node_iter = server_graph.getVertices().iterator(); node_iter.hasNext(); ) {
    		Vertex server = (Vertex) node_iter.next();
    		if (server.inDegree() == 0 || server.outDegree() == 0) {
    			continue;
    		}
    		E = Math.max(E, getServiceCurve(server).getLatency());
    	}
    	return E;
    }
    
    /** computes the value tau in [LeBoudec91] p. 111 */
    private double getMaximumPacketDelayVariation() {
    	double tau = 0.0;
    	for (Iterator node_iter = server_graph.getVertices().iterator(); node_iter.hasNext(); ) {
    		Vertex server = (Vertex) node_iter.next();
    		if (server.inDegree() == 0 || server.outDegree() == 0) {
    			continue;
    		}

    		double r = 0.0;
    		double b = 0.0;
        	for (Iterator iter = getAllIncomingFlows(server).iterator(); iter.hasNext(); ) {
        		Flow f = (Flow) iter.next();
        		r += f.ac.getSustainedRate();
        		b += f.ac.getTBBurst();
        	}
    		tau = Math.max(tau, b/r);
    	}
    	return tau;    	
    }

    /** computes the value u in [LeBoudec91] p. 111 */
    private double computeHighestU(double ny) {
    	double C_max = getMaxServerRate();
    	double u = 0.0;
    	for (Iterator node_iter = server_graph.getVertices().iterator(); node_iter.hasNext(); ) {
    		Vertex server = (Vertex) node_iter.next();
    		if (server.inDegree() == 0 || server.outDegree() == 0) {
    			continue;
    		}

    		double C_m = getIncomingFlowPeakRateBound(server, C_max);
    		double r_m = getServiceCurve(server).getSustainedRate();	
    		u = Math.max(u, (C_m - r_m) / (C_m - ny*r_m));
    	}
    	return u;    	
    }
    
    //---------------------------------------------------------------------
    // output bound computation
    //---------------------------------------------------------------------
    /**
     * Returns the output bound for a set of <code>outgoing_flows</code> from
     * the given <code>server</code>.<br/>
     * It does so by first iterating over all incoming edges of the server, summing
     * up the output bounds for all flows of interest (incoming flows contained also in
     * <code>outgoing_flows</code>) and all cross flows (incoming flows not
     * contained in <code>outgoing_flows</code>).<br/>
     * In a next step, the cross flows are factored out of the service curve to obtain
     * the effective service curve. Per default, blind multiplexing is assumed. To use
     * FIFO service instead you need to invoke <code>setUseFIFOService(true)</code> on
     * the analyser instance before computing any bounds.<br/>
     * The output bound is then computed using arrival and service curves only. Invoke
     * <code>setUseGamma(true)</code> on the analyser instance to globally activate the
     * additional use of the maximum service curve for output bound computation.
     * Finally, by invoking <code>setUseExtraGamma(true)</code> it is possible to convolve
     * the obtained output bound once more with the maximum service curve whose latency
     * has been removed before returning the result.
     * 
     * @param server the server at which the output bound is computed
     * @param outgoing_flows set of flows for which the output bound is computed
     * @return a <code>Curve</code> instance containing the output bound
     * @see NetworkAnalyser#setUseFifoService(boolean)
     * @see NetworkAnalyser#setUseGamma(boolean)
     * @see NetworkAnalyser#setUseExtraGamma(boolean)
     */
	public Curve computeOutputBound(Vertex server, Set outgoing_flows) {
    	Curve alpha_out = Curve.NULL_CURVE;

    	if (outgoing_flows.size() == 0) {
    		return alpha_out;
    	}
    	
		if (server.inDegree() == 0) {
			alpha_out = getSourceFlow(server, outgoing_flows);
		} else {
	    	Curve sum_alpha      = Curve.NULL_CURVE;
	    	Curve sum_alpha_excl = Curve.NULL_CURVE;
	    	Curve beta_effective = getServiceCurve(server);
	    	Curve gamma          = getMaxServiceCurve(server);
	    	
			for (Iterator pred_iter = server.getPredecessors().iterator(); pred_iter.hasNext(); ) {
				Vertex predecessor = (Vertex) pred_iter.next();

				Set incoming_flows = getIncomingFlows(predecessor, server);
				Set included_flows = SetUtils.getIntersection(incoming_flows, outgoing_flows);
				Set excluded_flows = SetUtils.getDifference(incoming_flows, included_flows);
				
				sum_alpha      = sum_alpha.add(computeOutputBound(predecessor, included_flows));
				sum_alpha_excl = sum_alpha_excl.add(computeOutputBound(predecessor, excluded_flows));
			}
			
	    	if (use_fifo_service) {
	    		beta_effective = Curve.computeFIFOServiceCurve(sum_alpha_excl, beta_effective);
	    	} else {
				beta_effective = beta_effective.sub(sum_alpha_excl).boundAtXAxis();
	    	}

	    	if (use_gamma) {
		    	alpha_out = sum_alpha.convolveAlmostConcave(gamma).deconvolve(beta_effective).makeArrivalCurve();
	    	} else {
		    	alpha_out = sum_alpha.deconvolve(beta_effective).makeArrivalCurve();
	    	}
	    	if (use_extra_gamma) {
	    		alpha_out = alpha_out.convolveAlmostConcave(gamma.removeLatency());
	    	}
		}
		return alpha_out;
    }

    //---------------------------------------------------------------------
    // fair queueing analysis
    //---------------------------------------------------------------------
	/**
	 * Performs a fair queueing analysis and returns the end-to-end service curve
	 * for a the given <code>flow_of_interest</code>.<br/>
	 * This analysis assumes that flows passing a server are isolated from each
	 * other by a scheduler that guarantees each flow a fair (and equal) share of
	 * the link.
	 * @param flow_of_interest the flow for which the end-to-end service curve is computed
	 * @return the end-to-end service curve
	 */
    public Curve performFairQueueingAnalysis(Flow flow_of_interest) {
    	Curve beta_endtoend = Curve.ZERO_DELAY_BURST_CURVE;
    	
		List path = flow_of_interest.path.subList(1, flow_of_interest.path.size());
		for (Iterator path_iter = path.iterator(); path_iter.hasNext(); ) {
			DirectedEdge e = (DirectedEdge) path_iter.next();
			
			Vertex server = e.getSource();

			double foi_r = flow_of_interest.ac.getSustainedRate();
			
			double sum_r = 0.0;
			for (Iterator flow_iter = getAllIncomingFlows(server).iterator(); flow_iter.hasNext(); ) {
				Flow f = (Flow) flow_iter.next();
				sum_r += f.ac.getSustainedRate();
			}
			
			Curve beta = getServiceCurve(server);
			double R = beta.getSustainedRate();
			double L = beta.getLatency();

			beta_endtoend = beta_endtoend.convolveConvex(Curve.createRateLatency(L, foi_r/sum_r * R));
		}

    	return beta_endtoend;
    }

    //-------------------------------------------------------------------------
    // total flow analysis
    //-------------------------------------------------------------------------
    /**
     * Performs a total flow analysis for the <code>flow_of_interest</code> and returns
     * the end-to-end delay bound. If a <code>Map</code> instance is passed via the
     * parameter <code>server_to_backlog_bound_map</code>, the map is filled with key-value-pairs
     * mapping servers (instances of <code>Vertex</code>) to backlog bounds (instances
     * of <code>Double</code>). If no such mapping is needed, pass <code>null</code> instead.<br/>
     * A total flow analysis iterates over the path of the flow of interest, summing up
     * the additional delay incurred by each server it traverses. Note that this form of
     * analysis is based on FIFO assumptions.
     * @param flow_of_interest the flow for which the bounds shall be computed
     * @param server_to_backlog_bound_map a <code>Map</code> instance that on return is filled with
     * &lt;server,backlog bound&gt;-pairs; pass <code>null</code> if no such mapping is needed
     * @return the end-to-end delay bound
     */
    public double performTotalFlowAnalysis(Flow flow_of_interest, Map server_to_backlog_bound_map) {
		double total_delay = 0.0;
		
		List path = flow_of_interest.path.subList(1, flow_of_interest.path.size());
		for (Iterator path_iter = path.iterator(); path_iter.hasNext(); ) {
			DirectedEdge e = (DirectedEdge) path_iter.next();
			
			Vertex server = e.getSource();
			
			Curve sum_alpha      = Curve.NULL_CURVE;
			Curve sum_alpha_excl = Curve.NULL_CURVE;
			for (Iterator pred_iter = server.getPredecessors().iterator(); pred_iter.hasNext(); ) {
				Vertex predecessor = (Vertex) pred_iter.next();

				Set incoming_flows = getIncomingFlows(predecessor, server);
				Set included_flows = SetUtils.getIntersection(incoming_flows, getFlowSet(e));
				Set excluded_flows = SetUtils.getDifference(incoming_flows, included_flows);
				
				sum_alpha      = sum_alpha.add(computeOutputBound(predecessor, included_flows));
				sum_alpha_excl = sum_alpha_excl.add(computeOutputBound(predecessor, excluded_flows));
			}

			Curve beta_effective = getServiceCurve(server);
			beta_effective = beta_effective.sub(sum_alpha_excl).boundAtXAxis();
			
			total_delay += Curve.getDelayBound(sum_alpha, beta_effective);
			if (server_to_backlog_bound_map != null) {
				server_to_backlog_bound_map.put(server,
						new Double(Curve.getBacklogBound(sum_alpha, beta_effective)));
			}
		}
    	return total_delay;
    }
    
    //---------------------------------------------------------------------
    // separated flow analysis
    //---------------------------------------------------------------------
    /**
     * Performs a separated flow analysis for the <code>flow_of_interest</code> and returns
     * the end-to-end service curve.<br/>
     * This analysis first blends out the flow of interest and then computes for each server
     * along this flow's path the effective service curve that results if all remaining flows
     * crossing this server receive their maximum amount of service. Then all effective
     * service curves are concatenated to receive the end-to-end service curve from the
     * perspective of the flow of interest.
     * @param flow_of_interest the flow for which the end-to-end service curve shall be computed
     * @return the end-to-end service curve
     */
    public Curve performSeparatedFlowAnalysis(Flow flow_of_interest) {
    	Curve beta_endtoend = Curve.ZERO_DELAY_BURST_CURVE;
    	
    	Set flows_to_analyse = new HashSet(Collections.singleton(flow_of_interest));

		List path = flow_of_interest.path.subList(1, flow_of_interest.path.size());
		for (Iterator path_iter = path.iterator(); path_iter.hasNext(); ) {
			DirectedEdge e = (DirectedEdge) path_iter.next();
			
			Vertex server = e.getSource();

			Curve sum_alpha_excl = Curve.NULL_CURVE;
			for (Iterator pred_iter = server.getPredecessors().iterator(); pred_iter.hasNext(); ) {
				Vertex predecessor = (Vertex) pred_iter.next();

				Set incoming_flows = getIncomingFlows(predecessor, server);
				Set excluded_flows = SetUtils.getDifference(incoming_flows, flows_to_analyse);
				
				sum_alpha_excl = sum_alpha_excl.add(computeOutputBound(predecessor, excluded_flows));
			}

	    	Curve beta_effective = getServiceCurve(server);
			if (use_fifo_service) {
	    		beta_effective = Curve.computeFIFOServiceCurve(sum_alpha_excl, beta_effective);
	    	} else {
				beta_effective = beta_effective.sub(sum_alpha_excl).boundAtXAxis();
	    	}
			beta_endtoend = beta_endtoend.convolveConvex(beta_effective);
		}
		return beta_endtoend;
	}
        
    //-------------------------------------------------------------------------
    // PMOO analysis
    //-------------------------------------------------------------------------
    /**
     * Performs a pay-multiplexing-only-once (PMOO) analysis for the
     * <code>flow_of_interest</code> and returns the end-to-end service curve.
     * @param flow_of_interest the flow for which the end-to-end service curve shall be computed
     * @return the PMOO end-to-end service curve
     */
    public Curve performPMOOAnalysis(Flow flow_of_interest) {
		this.rec_depth = -1;
		List path = flow_of_interest.path.subList(1, flow_of_interest.path.size());
    	return getPMOOServiceCurve(path, Collections.singleton(flow_of_interest));	
    }

    /**
     * Returns the PMOO service curve for a set of <code>flows_of_interest</code>.
     * This service curve is a concatenation of all servers on the given edge path
     * <code>path</code> that have outgoing edges contained in the path (i.e. the
     * server at which the path ends is not included).<br/>
     * For performance reasons, the method first computes various mappings of servers
     * to incoming and leaving flows. It then computes the output bounds of all
     * incoming flows. If there are rejoining flows (flows that were there before, then
     * left and now join again), these are substituted for fresh flows.
     * Finally, <code>computePMOOServiceCurve()</code> is invoked,
     * which does the real work of concatenating the service curves.<br/>
     * The output bounds are computed using either the standard
     * <code>computeOutputBound()</code> method that is also used for the total flow
     * and separated flow analyses or the new <code>computePMOOOutputBound()</code>
     * method that results in tighter bounds but also increases computation effort.
     * The latter recursively calls this method again. It is also possible to use
     * a mixture of both, using the PMOO output bound for the first levels of recursion
     * and the standard output bound for deeper levels. The recursion level from which
     * on the standard bound is used may be set via <code>setMaxRecursionDepthForPMOO()</code>.
     * The default is to always use PMOO. To always use the standard output bound
     * computation set the maximum recursion depth to 0.<br/>
     * Several techniques that significantly reduce the complexity of the service
     * curve computation are employed:
     * <ul>
     * <li><it>Flow Aggregation</it>: Flows having the same ingress and egress points are always
     * replaced by a single flow whose bound is the aggregate bound of all contained flows.</li>
     * <li><it>Flow Prolongation</it>: Flows having the same ingress but different egress points may
     * be aggregated by extending the shorter flow to the egress point of the longer flow.
     * This technique is only used if it comes for free, i.e. it does not impact the
     * output bound. This is the case if the prolongation does not affect the location of
     * the bottleneck server passed by any of the prolonged flows. To activate this technique,
     * use <code>setUseFlowProlongation()</code>.</li>
     * <li><it>Arrival Curve Approximation<it>: Approximates the arrival curves of
     * incoming flows by a different arrival curve with a lower number of linear segments. This
     * does have an effect on the output bound, but does not necessarily affect the delay bound.
     * To enable this technique, use <code>setUseArrivalCurveApproximation()</code>.</li>
     * </ul>
     * 
     * @param path the edge path along which service curves are concatenated (not including the
     * server at which the path ends)
     * @param flows_of_interest the set of flows of interest
     * @return the concatenated PMOO service curve
     * @see NetworkAnalyser#setMaxRecursionDepthForPMOO(int)
     * @see NetworkAnalyser#setUseFlowProlongation(boolean)
     * @see NetworkAnalyser#setUseArrivalCurveApproximation(boolean)
     */
    protected Curve getPMOOServiceCurve(List path, Set flows_of_interest) {
    	rec_depth++;
    	
    	List joining_flow_sets = getJoiningFlowSets(path, flows_of_interest);
    	
    	List rejoining_flow_sets = getRejoiningFlowSets(joining_flow_sets);
    	
    	Set all_interfering_flows = SetUtils.getUnion(joining_flow_sets);
    	
    	Map flow_egress_map = getFlowEgressMap(path, all_interfering_flows);
 
    	Map flow_bound_map = new HashMap();

    	int i = 0;
    	for (Iterator path_iter = path.iterator(); path_iter.hasNext(); i++) {
    		Vertex server = ((DirectedEdge) path_iter.next()).getSource();
    		
    		Set joining_flows = (Set) joining_flow_sets.get(i);
    		
    		Set rejoining_flows = (Set) rejoining_flow_sets.get(i);

    		Set aggregateable_flow_sets = getAggregateableFlowSets(path, i, joining_flows, flow_egress_map);
			for (Iterator flow_set_iter = aggregateable_flow_sets.iterator(); flow_set_iter.hasNext(); ) {
				Set aggregateable_flows = (Set) flow_set_iter.next();
				
				// search for subsets of the aggregate on all incoming links; computePBOOOutputBound()
				// will correctly handle the case of empty subsets on an incoming link
				Curve aggregate_bound = Curve.NULL_CURVE;
				for (Iterator inedge_iter = server.getInEdges().iterator(); inedge_iter.hasNext(); ) {
					DirectedEdge inedge = (DirectedEdge) inedge_iter.next();
					Set inedge_flows = getFlowSet(inedge);
					Set flow_set     = SetUtils.getIntersection(inedge_flows, aggregateable_flows);
					
					Curve alpha_excl;
					if (rec_depth < max_rec_depth_for_pmoo) {
						alpha_excl = computePMOOOutputBound(server, flow_set);
					} else {
						alpha_excl = computeOutputBound(inedge.getSource(), flow_set);						
					}
					aggregate_bound = aggregate_bound.add(alpha_excl);
				}
				
				Set rejoining_flows_in_aggregate = SetUtils.getIntersection(aggregateable_flows, rejoining_flows);
				if (rejoining_flows_in_aggregate.size() > 0) {
					int egress_server = findEgressPoint(path, i, (Flow) rejoining_flows_in_aggregate.iterator().next());
					createSubstituteFlows(rejoining_flows_in_aggregate,	egress_server, aggregateable_flows,
							all_interfering_flows, joining_flows, flow_egress_map);
				}
				
				Flow aggregate_flow;
				if (aggregateable_flows.size() == 1) {
					// store bounds for unaggregateable flows directly
					aggregate_flow = (Flow) aggregateable_flows.iterator().next();
				} else {
					aggregate_flow = aggregateFlows(aggregateable_flows,
							joining_flows, all_interfering_flows, flow_egress_map);
				}
				flow_bound_map.put(aggregate_flow, aggregate_bound);
			}
    	}
    	
    	if (use_arrival_curve_approximation) {
    		approximateArrivalCurves(all_interfering_flows, flow_bound_map);
    	}
    	if (use_flow_prolongation) {
    		prolongFlows(path, all_interfering_flows, joining_flow_sets, flow_egress_map, flow_bound_map);
    	}

    	rec_depth--;
    	return computePMOOServiceCurve(path, all_interfering_flows, joining_flow_sets,
    			flow_egress_map, flow_bound_map, rec_depth+1);
    }
    
    /**
     * Aggregates a set of flows by removing them from the internal management structures
     * and replacing it with a single dummy aggregate flow.
     * @param aggregateable_flows the flows to aggregate
     * @param joining_flows the set of flows with the same ingress as the aggregate
     * @param all_interfering_flows the set of all interfering flows
     * @param flow_egress_map a map of flows to egress servers
     * @return the aggregate flow. 
     */
    private Flow aggregateFlows(Set aggregateable_flows, Set joining_flows, Set all_interfering_flows, Map flow_egress_map) {
		Flow aggregate_flow;

		// backup egress server of the aggregateable flows
		Integer egress_server = (Integer) flow_egress_map.get(
				                   aggregateable_flows.iterator().next());
		
		// remove aggregateable flows from all local data structures...
		joining_flows.removeAll(aggregateable_flows);
		all_interfering_flows.removeAll(aggregateable_flows);
		for (Iterator flow_iter = aggregateable_flows.iterator(); flow_iter.hasNext(); ) {
			Flow f = (Flow) flow_iter.next();
			flow_egress_map.remove(f);
		}
		
		//...and create a new aggregate flow instead.
		aggregate_flow = Flow.createDummyFlow();
		joining_flows.add(aggregate_flow);
		all_interfering_flows.add(aggregate_flow);
		flow_egress_map.put(aggregate_flow, egress_server);
		
    	return aggregate_flow;
    }
    
    /**
     * Creates a list containing the joining flows sets for each server along the path.
     * Does not include the flows_of_interest.
     * @param path
     * @param flows_of_interest
     * @return a list containing sets of joining flows
     */
	private List getJoiningFlowSets(List path, Set flows_of_interest) {
		Set present_flows  = new HashSet(flows_of_interest);
		List joining_flow_sets = new ArrayList();
		for (Iterator path_iter = path.iterator(); path_iter.hasNext(); ) {
			DirectedEdge e = (DirectedEdge) path_iter.next();
			
			Vertex server = e.getSource();
			
			Set joining_flows = SetUtils.getDifference(getAllIncomingFlows(server), present_flows);
			joining_flow_sets.add(joining_flows);
			
			present_flows.addAll(joining_flows);
			
			Set leaving_flows = new HashSet();
			for (Iterator outedge_iter = server.getOutEdges().iterator(); outedge_iter.hasNext(); ) {
				DirectedEdge out_e = (DirectedEdge) outedge_iter.next();
				if (out_e == e) {
					// skip outgoing edge which is on path
					continue;
				}
				leaving_flows.addAll(getFlowSet(out_e));
			}
			present_flows = SetUtils.getDifference(present_flows, leaving_flows);
		}
		return joining_flow_sets;
	}
	
	
    /**
     * Creates a list containing a set of rejoining flows for each server along the path.
     * Rejoining flows are flows that have joined the flow of interest before, then
     * took a different path and finally join the flow of interest again.
     * @param joining_flow_sets the list of joining flows
     * @return a list containing sets of joining flows
     */
	private List getRejoiningFlowSets(List joining_flow_sets) {
		Set previously_joined_flows = new HashSet();
		
		List rejoining_flow_sets = new ArrayList(joining_flow_sets.size());
		for (int i = 0; i < joining_flow_sets.size(); i++) {
			Set rejoining_flows = SetUtils.getIntersection(previously_joined_flows, (Set) joining_flow_sets.get(i));
			rejoining_flow_sets.add(rejoining_flows);
			previously_joined_flows.addAll((Set) joining_flow_sets.get(i));
		}
		return rejoining_flow_sets;
	}
	
	/**
	 * Computes the map of flows to egress point (an Integer index into the list of servers
	 * on the path).
	 * @param path
	 * @param all_cross_traffic_flows
	 * @return the map
	 */
	private Map getFlowEgressMap(List path, Set all_cross_traffic_flows) {
		Map map = new HashMap();
		int i = 0;
    	for (Iterator path_iter = path.iterator(); path_iter.hasNext(); i++) {
			DirectedEdge e = (DirectedEdge) path_iter.next();
			
			Vertex server = e.getSource();
			
			Set outgoing_flows = new HashSet();
			for (Iterator outedge_iter = server.getOutEdges().iterator(); outedge_iter.hasNext(); ) {
				DirectedEdge out_e = (DirectedEdge) outedge_iter.next();
				if (out_e == e) {
					// skip outgoing edge which is on path
					continue;
				}
				outgoing_flows.addAll(getFlowSet(out_e));
			}
			outgoing_flows = SetUtils.getIntersection(outgoing_flows, all_cross_traffic_flows);
			for (Iterator flow_iter = outgoing_flows.iterator(); flow_iter.hasNext(); ) {
				Flow f = (Flow) flow_iter.next();
				if (!map.containsKey(f)) {
					map.put(f, new Integer(i)); // store only the first egress point of a flow
				}
			}			
		}
		for (Iterator flow_iter = all_cross_traffic_flows.iterator(); flow_iter.hasNext(); ) {
			Flow f = (Flow) flow_iter.next();
			if (!map.containsKey(f)) {
				map.put(f, new Integer(path.size()-1));
			}
		}			
		return map;
	}
	
	/**
	 * Creates a partition of the set of flows joining this server so that each
	 * partition contains flows with identical egress point.
	 * @param path
	 * @param server the ingress server
	 * @param joining_flows the set of flows joining at the ingress server
	 * @param flow_egress_map the flow-to-egress-point map
	 * @return a set of flow sets with same egress point
	 */
	private Set getAggregateableFlowSets(List path, int server, Set joining_flows, Map flow_egress_map) {
		Set aggregateable_flows = new HashSet();
		
		List remaining_flows = new ArrayList(joining_flows);
		while(remaining_flows.size() > 0) {
			Flow flow = (Flow) remaining_flows.remove(remaining_flows.size()-1);			
			
			Integer common_egress_server = (Integer) flow_egress_map.get(flow);
			if (common_egress_server == null || common_egress_server.intValue() < server) { // i.e. this flow is rejoining
				common_egress_server = new Integer(findEgressPoint(path, server, flow));
			}
		
			Set flow_set = new HashSet(Collections.singleton(flow));
			for (int i = remaining_flows.size()-1; i >= 0; i--) {
				flow = (Flow) remaining_flows.get(i);
				
				Integer egress_server = (Integer) flow_egress_map.get(flow);
				if (egress_server == null || egress_server.intValue() < server) { // i.e. this flow is rejoining
					egress_server = new Integer(findEgressPoint(path, server, flow));
				}
			
				if (egress_server.intValue() == common_egress_server.intValue()) {
					flow_set.add(flow);
					remaining_flows.remove(i);
				}
			}			
			aggregateable_flows.add(flow_set);
		}
		
		return aggregateable_flows;
	}

	/**
	 * Returns the index of the server at which the given flow exits the path. This point
	 * must be after the server specified by server_idx.
	 * @param path the path
	 * @param server_idx the index of the server where the flow joins the path
	 * @param flow the flow
	 * @return the index of the server in the path
	 */
	private int findEgressPoint(List path, int server_idx, Flow flow) {
		int i = 0;
    	for (Iterator path_iter = path.iterator(); path_iter.hasNext(); i++) {
    		DirectedEdge e = (DirectedEdge) path_iter.next();

    		if (i < server_idx) {
    			continue;
    		}

			Vertex server = e.getSource();
			for (Iterator outedge_iter = server.getOutEdges().iterator(); outedge_iter.hasNext(); ) {
				DirectedEdge out_e = (DirectedEdge) outedge_iter.next();
				if (out_e == e) {
					// skip outgoing edge which is on path
					continue;
				}
				if (getFlowSet(out_e).contains(flow)) {
					return i;
				}
			}
		}
		return path.size()-1;
	}

	/**
	 * Creates a fresh flow substitute for a rejoining flow and updates the respective
	 * management information.
	 * @param rejoining_flows_in_aggregate
	 * @param egress_server
	 * @param aggregateable_flows
	 * @param all_interfering_flows
	 * @param joining_flows
	 * @param flow_egress_map
	 */
	private void createSubstituteFlows(Set rejoining_flows_in_aggregate, int egress_server,
			Set aggregateable_flows, Set all_interfering_flows, Set joining_flows, Map flow_egress_map) {
		for (Iterator flow_iter = rejoining_flows_in_aggregate.iterator(); flow_iter.hasNext(); ) {
			Flow flow       = (Flow) flow_iter.next();
			Flow substitute = new Flow(null, null, null, null);
			aggregateable_flows.remove(flow);
			aggregateable_flows.add(substitute);
			all_interfering_flows.add(substitute);
			joining_flows.remove(flow);
			joining_flows.add(substitute);
			flow_egress_map.put(substitute, new Integer(egress_server));
		}
	}
	
	/**
	 * Computes the PMOO output bound for a set of <code>flows_of_interest</code>. The
	 * difference to the standard output bound method is that this method tries to
	 * compute tighter bounds by concatenating as many servers as possible using the
	 * PMOO approach. It does so by searching from <code>server</code> towards
	 * the sinks of the flows contained in <code>flows_of_interest</code> until it
	 * reaches the server where all these flows first meet each other (the "splitting point").
	 * It then concatenates all servers between the splitting point (inclusive) and
	 * <code>server</code> (exclusive).<br/>
	 * Note: The previous sentence means that the output bound is for the flows <b>before</b>
	 * <code>server</code> and not after the server as is the case with the standard
	 * computateOutputBound().
	 * @param server that all flows of interest flow into
	 * @param flows_of_interest the set of flows of interest
	 * @return the PMOO output bound
	 */
	protected Curve computePMOOOutputBound(Vertex server, Set flows_of_interest) {
    	Curve alpha_out = Curve.NULL_CURVE;

    	if (flows_of_interest.size() == 0) {
    		return alpha_out;
    	}
    	
    	Vertex split = findSplittingPoint(server, flows_of_interest);
    	if (split.equals(server)) {
    		throw new RuntimeException("could not find splitting point...");
    	}

    	Flow f = (Flow) flows_of_interest.iterator().next();
    	if (split.equals(f.source)) {
    		Curve alpha = getSourceFlow(f.source, flows_of_interest);
    		
       		List path = GraphUtils.getSubPath(f.path, f.source, server);
       		if (path.size() == 1) {
       			alpha_out = alpha;
       		} else {
        		path = path.subList(1, path.size());
        		
        		Curve beta      = getPMOOServiceCurve(path, flows_of_interest);
        		Curve gamma     = convolveMaxServiceCurves(path);
    	    	if (use_gamma) {
    	    		alpha_out = alpha.convolveAlmostConcave(gamma).deconvolve(beta).makeArrivalCurve();
    	    	} else {
    		    	alpha_out = alpha.deconvolve(beta).makeArrivalCurve();
    	    	}    		
    	    	if (use_extra_gamma) {
    	    		alpha_out = alpha_out.convolveAlmostConcave(gamma.removeLatency());
    	    	}
       		}
    	} else {
        	List path = GraphUtils.getSubPath(f.path, split, server);
        	
    		Curve alpha     = Curve.NULL_CURVE;
    		for (Iterator inedge_iter = split.getInEdges().iterator(); inedge_iter.hasNext(); ) {
				DirectedEdge inedge = (DirectedEdge) inedge_iter.next();
				
				Set inflows = getFlowSet(inedge);
				Set flows   = SetUtils.getIntersection(inflows, flows_of_interest);
				alpha = alpha.add(computePMOOOutputBound(split, flows));
    		}
    		Curve beta      = getPMOOServiceCurve(path, flows_of_interest);
    		Curve gamma     = convolveMaxServiceCurves(path);
    		if (use_gamma) {
	    		alpha_out = alpha.convolveAlmostConcave(gamma).deconvolve(beta).makeArrivalCurve();
	    	} else {
		    	alpha_out = alpha.deconvolve(beta).makeArrivalCurve();
	    	}    		
	    	if (use_extra_gamma) {
	    		alpha_out = alpha_out.convolveAlmostConcave(gamma.removeLatency());
	    	}
    	}
		return alpha_out;
	}

	/**
	 * Returns the server at which the flows in <code>flows_of_interest</code> first
	 * all meet each other (when viewed from the source). When viewed from
	 * <code>server</code> towards the sink, this is the last server where all flows
	 * are still together.
	 * @param server
	 * @param flows_of_interest
	 * @return the splitting point server
	 */
	protected Vertex findSplittingPoint(Vertex server, Set flows_of_interest) {
		Flow f = (Flow) flows_of_interest.iterator().next();
		
    	List path = GraphUtils.getSubPath(f.path, f.source, server);
    	
		for (Iterator path_iter = path.iterator(); path_iter.hasNext(); ) {
			DirectedEdge e = (DirectedEdge) path_iter.next();

			Set flow_set = getFlowSet(e);
			if (SetUtils.getIntersection(flow_set, flows_of_interest).size() == flows_of_interest.size()) {
				return e.getSource();
			}
		}
		return server;
	}

	/**
	 * Approximates arrival curves by ones with fewer number of segments. It removes
	 * segments in the order of ascending heights of a segment's projection onto the
	 * y-axis. It tries to continue until the average number of token buckets that
	 * each arrival curve can be decomposed into has reached the value set
	 * by <code>setTokenBucketCountTarget()</code> (defaults to 2.0).
	 * @param all_interfering_flows set of all flows interfering with the flow of interest
	 * @param flow_bound_map map of interfering flows to their arrival curve
	 */
	protected void approximateArrivalCurves(Set all_interfering_flows, Map flow_bound_map) {
		int segment_count = 0;
		
		List all_y_intervals = new ArrayList();
		for (Iterator flow_iter = all_interfering_flows.iterator(); flow_iter.hasNext(); ) {
			Flow f = (Flow) flow_iter.next();
			
			List y_intervals = ((Curve) flow_bound_map.get(f)).getYIntervals();
			all_y_intervals.addAll(y_intervals);
			
			segment_count += y_intervals.size();
		}
		Collections.sort(all_y_intervals);
		
		int segment_target = (int) Math.floor((token_bucket_count_target-1) * all_interfering_flows.size());
		if (segment_count <= segment_target) {
			return;
		}
		
		double cutoff = ((Double) all_y_intervals.get(segment_count - segment_target - 1)).doubleValue();
		
		for (Iterator flow_iter = all_interfering_flows.iterator(); flow_iter.hasNext(); ) {
			Flow f = (Flow) flow_iter.next();
			
			Curve bound = (Curve) flow_bound_map.get(f);
			segment_count -= bound.approximateSegments(cutoff);
		}
	}

	/**
	 * Tries to prolong flows having the same ingress but different egress points so that
	 * they may be aggregated. Tests first, whether this prolongation comes for free
	 * (i.e. without effect on the output bounds). If so, the internal management information
	 * is updated accordingly.
	 * @param path
	 * @param all_interfering_flows
	 * @param joining_flow_sets
	 * @param flow_egress_map
	 * @param flow_bound_map
	 */
	protected void prolongFlows(List path, Set all_interfering_flows, List joining_flow_sets,
			Map flow_egress_map, Map flow_bound_map) {

    	// compute residual rates: per server, min per flow
    	List server_residual_rates = new ArrayList(path.size());
    	List passing_flows = new ArrayList(path.size());
    	Map flow_min_residual_rates = new HashMap(flow_egress_map.size());
    	
    	Set present_flows = new HashSet();
    	
		int i = 0;
    	for (Iterator path_iter = path.iterator(); path_iter.hasNext(); i++) {
    		Vertex server = ((DirectedEdge) path_iter.next()).getSource();
			
    		// add incoming flows
			present_flows.addAll((Set) joining_flow_sets.get(i));

			passing_flows.add(new HashSet(present_flows));
			
			// calculate and store sum of rates of all passing flows
			double sum_r = 0.0;
			for (Iterator flow_iter = present_flows.iterator(); flow_iter.hasNext(); ) {
				Flow f = (Flow) flow_iter.next();

				Curve bound      = (Curve) flow_bound_map.get(f);
				sum_r += bound.getSustainedRate();
			}

			// compute residual rate
			double server_rate = getServiceCurve(server).getSustainedRate();
			server_residual_rates.add(new Double(server_rate - sum_r));
			
			// update minimum residual rate for each flow
			for (Iterator flow_iter = present_flows.iterator(); flow_iter.hasNext(); ) {
				Flow f = (Flow) flow_iter.next();
				
				if (!flow_min_residual_rates.containsKey(f)) {
					flow_min_residual_rates.put(f, new Double(server_rate - sum_r));
				} else {
					double current_R = ((Double) flow_min_residual_rates.get(f)).doubleValue();
					flow_min_residual_rates.put(f, new Double(Math.min(current_R, server_rate - sum_r)));
				}
			}
			
			// remove all outgoing flows from the set of present flows
			for (Iterator flow_iter = present_flows.iterator(); flow_iter.hasNext(); ) {
				Flow f = (Flow) flow_iter.next();
				
				int egress_server = ((Integer) flow_egress_map.get(f)).intValue();
				if (egress_server <= i) {
					flow_iter.remove();
				}
			}			
		}

    	
    	// test for prolongable flows
		for (i = 0; i < path.size(); i++) {
			Set joining_flows = (Set) joining_flow_sets.get(i);
			
			List list_of_joining_flows = new ArrayList(joining_flows);
			Collections.sort(list_of_joining_flows, new MapValueComparator(flow_egress_map));
			
			for (int j = 0; j < list_of_joining_flows.size()-1; j++) {
				Flow f1 = (Flow) list_of_joining_flows.get(j);
				Flow f2 = (Flow) list_of_joining_flows.get(j+1);
				if (isFreeProlongation(f1, f2, server_residual_rates, passing_flows,
						flow_min_residual_rates, flow_egress_map, flow_bound_map)) {
					
					flow_egress_map.put(f1, flow_egress_map.get(f2));
					
					Curve aggregate_bound = Curve.NULL_CURVE;
					aggregate_bound = aggregate_bound.add((Curve) flow_bound_map.get(f1));
					aggregate_bound = aggregate_bound.add((Curve) flow_bound_map.get(f2));
					flow_bound_map.remove(f1);
					flow_bound_map.remove(f2);
					
					Set aggregateable_flows = new HashSet();
					aggregateable_flows.add(f1);
					aggregateable_flows.add(f2);
					Flow aggregate_flow = aggregateFlows(aggregateable_flows,
							joining_flows, all_interfering_flows, flow_egress_map);
					
					flow_bound_map.put(aggregate_flow, aggregate_bound);
					break;
				}
			}
		}
	}
	
	/**
	 * Tests whether the prolongation of a flow is for free, i.e. it does not change
	 * the output bound. This is the case if the prolongation does not change the
	 * bottleneck server which is the server with the lowest residual rate encountered
	 * by any of the flows on the subpath that the prolonged flow shares.
	 * @param short_flow
	 * @param long_flow
	 * @param server_residual_rates
	 * @param passing_flow_sets
	 * @param flow_min_residual_rates
	 * @param flow_egress_map
	 * @param flow_bound_map
	 * @return whether the prolongation is for free or not
	 */
	protected boolean isFreeProlongation(Flow short_flow, Flow long_flow, List server_residual_rates,
			List passing_flow_sets, Map flow_min_residual_rates, Map flow_egress_map, Map flow_bound_map) {
		int egress_short = ((Integer) flow_egress_map.get(short_flow)).intValue();
		int egress_long  = ((Integer) flow_egress_map.get(long_flow)).intValue();
		
		double additional_rate = ((Curve) flow_bound_map.get(short_flow)).getSustainedRate();
		for (int i = egress_short+1; i <= egress_long; i++) {
			double residual_rate = ((Double) server_residual_rates.get(i)).doubleValue();

			Set passing_flows = (Set) passing_flow_sets.get(i);
			for (Iterator flow_iter = passing_flows.iterator(); flow_iter.hasNext(); ) {
				Flow f = (Flow) flow_iter.next();
				
				double flows_min_encountered_rate = ((Double) flow_min_residual_rates.get(f)).doubleValue();
				if (residual_rate - additional_rate < flows_min_encountered_rate) {
					return false;
				}				
			}
		}
		return true;
	}
	
    /**
     * Concatenates the service curves along the given path <code>path</code> according to
     * the PMOO approach and returns the result.<br/>
     * It first decomposes all arrival curves (service curves) into token buckets (rate
     * latency curves), enumerates over all combinations of token buckets and rate latency curves,
     * and calls <code>computePartialPMOOServiceCurve()</code> for each combination. The
     * total PMOO service curve is the maximum of all partial service curves.
     * @param path edge path traversing the servers whose service curves shall be concatenated
     * (excluding the one at which the path ends).
     * @param all_interfering_flows all flows that cross one of the servers on the path and
     * do not belong to the flows of interest
     * @param joining_flow_sets a list of sets with the flows that join at each server
     * @param flow_egress_map map of flows to egress points (integer index into the path)
     * @param flow_bound_map map of flows to their arrival curves
     * @param rec_depth the current recursion level (optional)
     * @return the PMOO service curve
     */
    protected Curve computePMOOServiceCurve(List path, Set all_interfering_flows,
    		List joining_flow_sets, Map flow_egress_map, Map flow_bound_map, int rec_depth) {

    	int combinations_to_test = 1;
    	int tb_component_count = 0;
    	int rl_component_count = 0;
    	
    	// create a flow-->tb_iter map
    	Map flow_tb_iter_map = new HashMap();
		for (Iterator flow_iter = all_interfering_flows.iterator(); flow_iter.hasNext(); ) {
			Flow f = (Flow) flow_iter.next();
			flow_tb_iter_map.put(f, new Integer(0));
			combinations_to_test *= ((Curve) flow_bound_map.get(f)).getTBComponentCount();
	    	tb_component_count   += ((Curve) flow_bound_map.get(f)).getTBComponentCount();
		}
		// create a list of rl_iters
		Curve[] service_curves = new Curve[path.size()];
    	int[] server_rl_iters  = new int[path.size()];
    	int[] server_rl_counts = new int[path.size()];
    	int i = 0;
		for (Iterator path_iter = path.iterator(); path_iter.hasNext(); i++) {
			Vertex server = ((DirectedEdge) path_iter.next()).getSource();
			Curve service_curve = getServiceCurve(server);
			service_curves[i]   = service_curve;
			server_rl_iters[i]  = 0;
			server_rl_counts[i] = service_curve.getRLComponentCount();
			combinations_to_test *= service_curve.getRLComponentCount();
			rl_component_count   += service_curve.getRLComponentCount();
		}
    	
		Curve beta_total = Curve.NULL_CURVE;

    	boolean more_combinations = true;
    	while (more_combinations) {
    		// compute service curve for this combination
    		Curve beta = computePartialPMOOServiceCurve(service_curves, all_interfering_flows, joining_flow_sets,
			        flow_egress_map, flow_bound_map, flow_tb_iter_map, server_rl_iters);
    		if (beta != Curve.NULL_CURVE) {
    			beta_total = beta_total.max(beta);
    		}
    		
    		// first check whether there are more combinations of flow TBs
    		more_combinations = false;
    		for (Iterator flow_iter = flow_tb_iter_map.keySet().iterator(); flow_iter.hasNext(); ) {
    			Flow f = (Flow) flow_iter.next();

    			Curve f_bound = (Curve) flow_bound_map.get(f);
    			
    			i = ((Integer) flow_tb_iter_map.get(f)).intValue();
    			if (i+1 < f_bound.getTBComponentCount()) {
    				flow_tb_iter_map.put(f, new Integer(i+1));
    				more_combinations = true;
    				break;
    			} else {
    				flow_tb_iter_map.put(f, new Integer(0));
    			}
    		}
    		
    		// if not, check whether there are more combinations of server RLs
    		if (!more_combinations) {
    	    	for (i = 0; i < server_rl_iters.length; i++) {
        			int j = server_rl_iters[i];
        			if (j+1 < server_rl_counts[i]) {
        				server_rl_iters[i] = j;
        				more_combinations = true;
        				break;
        			} else {
        				server_rl_iters[i] = 0;
        			}
        		}    			
    		}
    	}
    	
		return beta_total;
    }
    
    /**
     * Calculates the partial PMOO service curve for the given flow set by combining
     * all servers having an outgoing edge contained in the given edge-path.
     * For each flow considers only one of its token bucket components (selected
     * via the flow_tb_iter_map) and for each service curve considers only one
     * rate latency curve (selected via the server_rl_iters).
     * @return a partial PMOO service curve
     */
    protected Curve computePartialPMOOServiceCurve(Curve[] service_curves, Set all_interfering_flows, List joining_flow_sets,
    		Map flow_egress_map, Map flow_bound_map, Map flow_tb_iter_map, int[] server_rl_iters) {

    	double T = 0.0;
    	double R = Double.POSITIVE_INFINITY;
    	double sum_bursts = 0.0;
    	double sum_latencyterms = 0.0;
    	
    	Set present_flows = new HashSet();
    	for (int i = 0; i < service_curves.length; i++) {
    		// add incoming flows
			present_flows.addAll((Set) joining_flow_sets.get(i));
			
			Curve current_rl = service_curves[i].getRLComponent(server_rl_iters[i]);

			// sum up latencies
			T += current_rl.getLatency();
			
			// calculate and store sum of rates of all passing flows
			double sum_r = 0.0;
			for (Iterator flow_iter = present_flows.iterator(); flow_iter.hasNext(); ) {
				Flow f = (Flow) flow_iter.next();

				Curve bound      = (Curve) flow_bound_map.get(f);
				Curve current_tb = bound.getTBComponent(((Integer) flow_tb_iter_map.get(f)).intValue());
				sum_r += current_tb.getSustainedRate();
			}
			
			// update latency terms
			sum_latencyterms += sum_r * current_rl.getLatency(); 

			// compute residual rate; update min
			double Ri = current_rl.getSustainedRate() - sum_r;
			if (Ri <= 0.0) {
				return Curve.NULL_CURVE;
			}
			R = Math.min(R, Ri);
			
			// remove all outgoing flows from the set of present flows
			for (Iterator flow_iter = present_flows.iterator(); flow_iter.hasNext(); ) {
				Flow f = (Flow) flow_iter.next();
				
				int egress_server = ((Integer) flow_egress_map.get(f)).intValue();
				if (egress_server <= i) {
					flow_iter.remove();
				}
			}			
		}
    	
    	// compute sum of bursts
		for (Iterator flow_iter = all_interfering_flows.iterator(); flow_iter.hasNext(); ) {
			Flow f = (Flow) flow_iter.next();
			
			Curve bound      = (Curve) flow_bound_map.get(f);
			Curve current_tb = bound.getTBComponent(((Integer) flow_tb_iter_map.get(f)).intValue());			
			sum_bursts += current_tb.getTBBurst();
		}    	
    	
		T += (sum_bursts + sum_latencyterms) / R;
		
		return Curve.createRateLatency(T, R);
    }
}
